# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from phono3py import Phono3py
from phono3py.file_IO import write_disp_fc3_yaml, write_disp_fc2_yaml
from phono3py.interface.calculator import (
    get_additional_info_to_write_fc2_supercells,
    get_additional_info_to_write_supercells,
    get_default_displacement_distance)
from phonopy.interface.calculator import write_supercells_with_displacements


def create_phono3py_supercells(cell_info,
                               settings,
                               symprec,
                               output_filename=None,
                               interface_mode='vasp',
                               log_level=1):
    optional_structure_info = cell_info['optional_structure_info']

    if settings.displacement_distance is None:
        distance = get_default_displacement_distance(interface_mode)
    else:
        distance = settings.displacement_distance
    phono3py = Phono3py(
        cell_info['unitcell'],
        cell_info['supercell_matrix'],
        primitive_matrix=cell_info['primitive_matrix'],
        phonon_supercell_matrix=cell_info['phonon_supercell_matrix'],
        is_symmetry=settings.is_symmetry,
        symprec=symprec,
        calculator=interface_mode)
    phono3py.generate_displacements(
        distance=distance,
        cutoff_pair_distance=settings.cutoff_pair_distance,
        is_plusminus=settings.is_plusminus_displacement,
        is_diagonal=settings.is_diagonal_displacement)

    if log_level:
        print('')
        print("Displacement distance: %s" % distance)

    if output_filename is None:
        filename = 'disp_fc3.yaml'
    else:
        filename = 'disp_fc3.' + output_filename + '.yaml'
    num_disps, num_disp_files = write_disp_fc3_yaml(phono3py.dataset,
                                                    phono3py.supercell,
                                                    filename=filename)
    ids = []
    disp_cells = []
    for i, cell in enumerate(phono3py.supercells_with_displacements):
        if cell is not None:
            ids.append(i + 1)
            disp_cells.append(cell)

    additional_info = get_additional_info_to_write_supercells(
        interface_mode, phono3py.supercell_matrix)
    write_supercells_with_displacements(interface_mode,
                                        phono3py.supercell,
                                        disp_cells,
                                        optional_structure_info,
                                        displacement_ids=ids,
                                        zfill_width=5,
                                        additional_info=additional_info)

    if log_level:
        print("Number of displacements: %d" % num_disps)
        if settings.cutoff_pair_distance is not None:
            print("Cutoff distance for displacements: %s" %
                  settings.cutoff_pair_distance)
            print("Number of displacement supercell files created: %d" %
                  num_disp_files)

    if phono3py.phonon_supercell_matrix is not None:
        if output_filename is None:
            filename = 'disp_fc2.yaml'
        else:
            filename = 'disp_fc2.' + output_filename + '.yaml'

        num_disps = write_disp_fc2_yaml(phono3py.phonon_dataset,
                                        phono3py.phonon_supercell,
                                        filename=filename)
        additional_info = get_additional_info_to_write_fc2_supercells(
            interface_mode, phono3py.phonon_supercell_matrix)
        write_supercells_with_displacements(
            interface_mode,
            phono3py.supercell,
            phono3py.phonon_supercells_with_displacements,
            optional_structure_info,
            zfill_width=5,
            additional_info=additional_info)

        if log_level:
            print("Number of displacements for special fc2: %d" % num_disps)

    return phono3py



# Copyright (C) 2016 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from phonopy.interface.phonopy_yaml import PhonopyYaml
import numpy as np


class Phono3pyYaml(PhonopyYaml):

    command_name = "phono3py"
    default_filenames = ("phono3py_disp.yaml", "phono3py.yaml")
    default_settings = {'force_sets': False,
                        'displacements': True,
                        'force_constants': False,
                        'born_effective_charge': True,
                        'dielectric_constant': True}

    def __init__(self,
                 configuration=None,
                 calculator=None,
                 physical_units=None,
                 settings=None):

        self.configuration = None
        self.calculator = None
        self.physical_units = None
        self.settings = None

        self.unitcell = None
        self.primitive = None
        self.supercell = None
        self.dataset = None
        self.supercell_matrix = None
        self.primitive_matrix = None
        self.nac_params = None
        self.force_constants = None

        self.symmetry = None  # symmetry of supercell
        self.s2p_map = None
        self.u2p_map = None
        self.frequency_unit_conversion_factor = None
        self.version = None

        #
        # phono3py only
        #
        # With DIM_FC2 given
        self.phonon_supercell_matrix = None
        self.phonon_dataset = None
        self.phonon_supercell = None
        self.phonon_primitive = None

        self._yaml = None

        super(Phono3pyYaml, self).__init__(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
            settings=settings)

    def set_phonon_info(self, phono3py):
        super(Phono3pyYaml, self).set_phonon_info(phono3py)
        self.phonon_supercell_matrix = phono3py.phonon_supercell_matrix
        self.phonon_dataset = phono3py.phonon_dataset
        self.phonon_primitive = phono3py.phonon_primitive
        self.phonon_supercell = phono3py.phonon_supercell

    def parse(self):
        super(Phono3pyYaml, self).parse()
        self._parse_fc3_dataset()

    def _load(self, fp):
        super(Phono3pyYaml, self)._load(fp)

    def _parse_all_cells(self):
        """Parse all cells

        This method override PhonopyYaml._parse_all_cells.

        """

        super(Phono3pyYaml, self)._parse_all_cells()
        if 'phonon_primitive_cell' in self._yaml:
            self.phonon_primitive = self._parse_cell(
                self._yaml['phonon_primitive_cell'])
        if 'phonon_supercell' in self._yaml:
            self.phonon_supercell = self._parse_cell(
                self._yaml['phonon_supercell'])
        if 'phonon_supercell_matrix' in self._yaml:
            self.phonon_supercell_matrix = np.array(
                self._yaml['phonon_supercell_matrix'], dtype='intc', order='C')

    def _parse_dataset(self):
        """Parse phonon_dataset

        This method override PhonopyYaml._parse_dataset.

        """

        self.phonon_dataset = self._get_dataset(self.phonon_supercell)

    def _parse_fc3_dataset(self):
        dataset = None
        if 'displacement_pairs' in self._yaml:
            disp = self._yaml['displacement_pairs'][0]
            if type(disp) is dict:  # type1
                dataset = self._parse_forces_fc3_type1(len(self.supercell))
            elif type(disp) is list:  # type2
                if 'displacement' in disp[0]:
                    dataset = self._parse_force_sets_type2()
        if 'displacement_pair_info' in self._yaml:
            info_yaml = self._yaml['displacement_pair_info']
            if 'cutoff_pair_distance' in info_yaml:
                dataset['cutoff_distance'] = info_yaml['cutoff_pair_distance']
            if 'duplicated_supercell_ids' in info_yaml:
                dataset['duplicates'] = info_yaml['duplicated_supercell_ids']
        self.dataset = dataset

    def _parse_forces_fc3_type1(self, natom):
        dataset = {'natom': natom, 'first_atoms': []}
        for d1 in self._yaml['displacement_pairs']:
            data1 = {
                'number': d1['atom'] - 1,
                'displacement': np.array(d1['displacement'], dtype='double'),
                'second_atoms': []}
            if 'forces' in d1:
                data1['forces'] = np.array(d1['forces'],
                                           dtype='double', order='C')
            d2_list = d1.get('paired_with')
            if d2_list is None:  # backward compatibility
                d2_list = d1.get('second_atoms')
            for d2 in d2_list:
                if 'forces' in d2:
                    data1['second_atoms'].append(
                        {'number': d2['atom'] - 1,
                         'displacement': np.array(d2['displacement'],
                                                  dtype='double'),
                         'forces': np.array(d2['forces'],
                                            dtype='double', order='C'),
                         'id': d2['displacement_id'],
                         'pair_distance': d2['pair_distance']})
                else:
                    disps = [{'number': d2['atom'] - 1,
                              'displacement': np.array(disp, dtype='double')}
                             for disp in d2['displacements']]
                    if 'pair_distance' in d2:
                        for d2_dict in disps:
                            d2_dict['pair_distance'] = d2['pair_distance']
                    if 'included' in d2:
                        for d2_dict in disps:
                            d2_dict['included'] = d2['included']
                    if 'displacement_ids' in d2:
                        for disp_id, d2_dict in zip(
                                d2['displacement_ids'], disps):
                            d2_dict['id'] = disp_id
                    data1['second_atoms'] += disps
            dataset['first_atoms'].append(data1)
        return dataset

    def _cell_info_yaml_lines(self):
        """Get YAML lines for information of cells

        This method override PhonopyYaml._cell_info_yaml_lines.

        """

        lines = super(Phono3pyYaml, self)._cell_info_yaml_lines()
        lines += self._supercell_matrix_yaml_lines(
            self.phonon_supercell_matrix, "phonon_supercell_matrix")
        lines += self._primitive_yaml_lines(self.phonon_primitive,
                                            "phonon_primitive_cell")
        lines += self._phonon_supercell_yaml_lines()
        return lines

    def _phonon_supercell_matrix_yaml_lines(self):
        lines = []
        if self.phonon_supercell_matrix is not None:
            lines.append("phonon_supercell_matrix:")
            for v in self.supercell_matrix:
                lines.append("- [ %3d, %3d, %3d ]" % tuple(v))
            lines.append("")
        return lines

    def _phonon_supercell_yaml_lines(self):
        lines = []
        if self.phonon_supercell is not None:
            s2p_map = getattr(self.phonon_primitive, 's2p_map', None)
            lines += self._cell_yaml_lines(
                self.phonon_supercell, "phonon_supercell", s2p_map)
            lines.append("")
        return lines

    def _nac_yaml_lines(self):
        """Get YAML lines for parameters of non-analytical term correction

        This method override PhonopyYaml._nac_yaml_lines.

        """

        if self.phonon_primitive is not None:
            return self._nac_yaml_lines_given_symbols(
                self.phonon_primitive.symbols)
        else:
            return self._nac_yaml_lines_given_symbols(
                self.primitive.symbols)

    def _displacements_yaml_lines(self, with_forces=False):
        """Get YAML lines for phonon_dataset and dataset.

        This method override PhonopyYaml._displacements_yaml_lines.
        PhonopyYaml._displacements_yaml_lines_2types is written
        to be also used by Phono3pyYaml.

        """

        lines = []
        if self.phonon_supercell_matrix is not None:
            lines += self._displacements_yaml_lines_2types(
                self.phonon_dataset, with_forces=with_forces)
        lines += self._displacements_yaml_lines_2types(
            self.dataset, with_forces=with_forces)
        return lines

    def _displacements_yaml_lines_type1(self, dataset, with_forces=False):
        """Get YAML lines for type1 phonon_dataset and dataset.

        This method override PhonopyYaml._displacements_yaml_lines_type1.
        PhonopyYaml._displacements_yaml_lines_2types calls
        Phono3pyYaml._displacements_yaml_lines_type1.

        """

        id_offset = len(dataset['first_atoms'])

        if 'second_atoms' in dataset['first_atoms'][0]:
            lines = ["displacement_pairs:"]
        else:
            lines = ["displacements:"]
        for i, d in enumerate(dataset['first_atoms']):
            lines.append("- atom: %4d" % (d['number'] + 1))
            lines.append("  displacement:")
            lines.append("    [ %19.16f, %19.16f, %19.16f ]"
                         % tuple(d['displacement']))
            id_num = i + 1
            if 'id' in d:
                assert id_num == d['id']
            lines.append("  displacement_id: %d" % id_num)
            if with_forces and 'forces' in d:
                lines.append("  forces:")
                for v in d['forces']:
                    lines.append(
                        "  - [ %19.16f, %19.16f, %19.16f ]" % tuple(v))
            if 'second_atoms' in d:
                ret_lines, id_offset = self._second_displacements_yaml_lines(
                    d['second_atoms'], id_offset, with_forces=with_forces)
                lines += ret_lines
        lines.append("")

        if 'second_atoms' in dataset['first_atoms'][0]:
            n_single = len(dataset['first_atoms'])
            n_pair = 0
            n_included = 0
            for d1 in dataset['first_atoms']:
                n_d2 = len(d1['second_atoms'])
                n_pair += n_d2
                for d2 in d1['second_atoms']:
                    if 'included' not in d2:
                        n_included += 1
                    elif d2['included']:
                        n_included += 1

            lines.append("displacement_pair_info:")
            if 'cutoff_distance' in dataset:
                lines.append("  cutoff_pair_distance: %11.8f"
                             % dataset['cutoff_distance'])
            lines.append("  number_of_singles: %d" % n_single)
            lines.append("  number_of_pairs: %d" % n_pair)
            if 'cutoff_distance' in dataset:
                lines.append("  number_of_pairs_in_cutoff: %d"
                             % n_included)

            if 'duplicates' in dataset and dataset['duplicates']:
                lines.append("  duplicated_supercell_ids: "
                             "# 0 means perfect supercell")
                for i in dataset['duplicates']:
                    # id-i and id-j give the same displacement pairs.
                    j = dataset['duplicates'][i]
                    # i can be str depending on data source.
                    lines.append("  - [ %d, %d ]" % (i, j))
            lines.append("")

        return lines

    def _second_displacements_yaml_lines(self,
                                         dataset2,
                                         id_offset,
                                         with_forces=False):
        lines = []
        id_num = id_offset
        # lines.append("  second_atoms:")
        lines.append("  paired_with:")
        numbers = np.array([d['number'] for d in dataset2])
        unique_numbers = np.unique(numbers)
        for i in unique_numbers:
            indices_eq_i = np.sort(np.where(numbers == i)[0])
            if with_forces and 'forces' in dataset2[indices_eq_i[0]]:
                for j in indices_eq_i:
                    id_num += 1
                    lines.append("  - atom: %4d" % (i + 1))
                    lines.append("    pair_distance: %.8f"
                                 % dataset2[j]['pair_distance'])
                    lines.append("    displacement:")
                    lines.append("      [ %19.16f, %19.16f, %19.16f ]"
                                 % tuple(dataset2[j]['displacement']))

                    if 'id' in dataset2[j]:
                        assert dataset2[j]['id'] == id_num
                        lines.append("    displacement_id: %d" % id_num)

                    lines.append("    forces:")
                    for v in dataset2[j]['forces']:
                        lines.append(
                            "    - [ %19.16f, %19.16f, %19.16f ]" % tuple(v))
            else:
                lines.append("  - atom: %4d" % (i + 1))
                lines.append("    pair_distance: %.8f"
                             % dataset2[indices_eq_i[0]]['pair_distance'])
                if 'included' in dataset2[indices_eq_i[0]]:
                    included = dataset2[indices_eq_i[0]]['included']
                    lines.append("    included: %s"
                                 % ("true" if included else "false"))
                disp_ids = []
                lines.append("    displacements:")
                for j in indices_eq_i:
                    id_num += 1
                    d = tuple(dataset2[j]['displacement'])
                    lines.append("    - [ %19.16f, %19.16f, %19.16f ]" % d)
                    if 'id' in dataset2[j]:
                        assert dataset2[j]['id'] == id_num
                        disp_ids.append(dataset2[j]['id'])
                if disp_ids:
                    lines.append("    displacement_ids: [ %s ]"
                                 % ', '.join(["%d" % j for j in disp_ids]))

        return lines, id_num



# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np


def run_phonon_solver_c(dm,
                        frequencies,
                        eigenvectors,
                        phonon_done,
                        grid_points,
                        grid_address,
                        mesh,
                        frequency_conversion_factor,
                        nac_q_direction,  # in reduced coordinates
                        lapack_zheev_uplo,
                        verbose=False):
    import phono3py._phono3py as phono3c

    (svecs,
     multiplicity,
     masses,
     rec_lattice,  # column vectors
     positions,
     born,
     nac_factor,
     dielectric) = _extract_params(dm)

    if dm.is_nac() and dm.nac_method == 'gonze':
        gonze_nac_dataset = dm.Gonze_nac_dataset
        if gonze_nac_dataset[0] is None:
            dm.make_Gonze_nac_dataset()
            gonze_nac_dataset = dm.Gonze_nac_dataset
        (gonze_fc,  # fc where the dipole-diple contribution is removed.
         dd_q0,     # second term of dipole-dipole expression.
         G_cutoff,  # Cutoff radius in reciprocal space. This will not be used.
         G_list,    # List of G points where d-d interactions are integrated.
         Lambda) = gonze_nac_dataset  # Convergence parameter
        fc = gonze_fc
    else:
        positions = None
        dd_q0 = None
        G_list = None
        Lambda = 0
        fc = dm.force_constants

    assert grid_points.dtype == 'uintp'
    assert grid_points.flags.c_contiguous

    fc_p2s, fc_s2p = _get_fc_elements_mapping(dm, fc)
    phono3c.phonons_at_gridpoints(
        frequencies,
        eigenvectors,
        phonon_done,
        grid_points,
        grid_address,
        np.array(mesh, dtype='intc'),
        fc,
        svecs,
        multiplicity,
        positions,
        masses,
        fc_p2s,
        fc_s2p,
        frequency_conversion_factor,
        born,
        dielectric,
        rec_lattice,
        nac_q_direction,
        nac_factor,
        dd_q0,
        G_list,
        Lambda,
        lapack_zheev_uplo)


def run_phonon_solver_py(grid_point,
                         phonon_done,
                         frequencies,
                         eigenvectors,
                         grid_address,
                         mesh,
                         dynamical_matrix,
                         frequency_conversion_factor,
                         lapack_zheev_uplo):
    gp = grid_point
    if phonon_done[gp] == 0:
        phonon_done[gp] = 1
        q = grid_address[gp].astype('double') / mesh
        dynamical_matrix.run(q)
        dm = dynamical_matrix.dynamical_matrix
        eigvals, eigvecs = np.linalg.eigh(dm, UPLO=lapack_zheev_uplo)
        eigvals = eigvals.real
        frequencies[gp] = (np.sqrt(np.abs(eigvals)) * np.sign(eigvals)
                           * frequency_conversion_factor)
        eigenvectors[gp] = eigvecs


def _extract_params(dm):
    svecs, multiplicity = dm.primitive.get_smallest_vectors()
    masses = np.array(dm.primitive.masses, dtype='double')
    rec_lattice = np.array(np.linalg.inv(dm.primitive.cell),
                           dtype='double', order='C')
    positions = np.array(dm.primitive.positions, dtype='double', order='C')
    if dm.is_nac():
        born = dm.born
        nac_factor = dm.nac_factor
        dielectric = dm.dielectric_constant
    else:
        born = None
        nac_factor = 0
        dielectric = None

    return (svecs,
            multiplicity,
            masses,
            rec_lattice,
            positions,
            born,
            nac_factor,
            dielectric)


def _get_fc_elements_mapping(dm, fc):
    p2s_map = dm.primitive.p2s_map
    s2p_map = dm.primitive.s2p_map
    if fc.shape[0] == fc.shape[1]:  # full fc
        fc_p2s = p2s_map
        fc_s2p = s2p_map
    else:  # compact fc
        primitive = dm.primitive
        p2p_map = primitive.p2p_map
        s2pp_map = np.array([p2p_map[s2p_map[i]] for i in range(len(s2p_map))],
                            dtype='intc')
        fc_p2s = np.arange(len(p2s_map), dtype='intc')
        fc_s2p = s2pp_map

    return fc_p2s, fc_s2p



# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
from phonopy.cui.phonopy_argparse import fix_deprecated_option_names


def get_parser(fc_symmetry=False,
               is_nac=False,
               load_phono3py_yaml=False):
    deprecated = fix_deprecated_option_names(sys.argv)
    import argparse
    from phonopy.interface.calculator import add_arguments_of_calculators
    from phono3py.interface.calculator import calculator_info

    parser = argparse.ArgumentParser(
        description="Phono3py command-line-tool")

    add_arguments_of_calculators(parser, calculator_info)

    parser.add_argument(
        "--alm", dest="use_alm", action="store_true", default=False,
        help=("Use ALM for generating 2nd and 3rd force constants "
              "in one fitting"))
    parser.add_argument(
        "--amplitude", dest="displacement_distance", type=float, default=None,
        help="Distance of displacements")
    parser.add_argument(
        "--ave-pp", dest="use_ave_pp", action="store_true", default=False,
        help="Use averaged ph-ph interaction")
    parser.add_argument(
        "--band", nargs='+', dest="band_paths", default=None,
        help="Band structure paths calculated for Gruneisen parameter")
    parser.add_argument(
        "--band-points", dest="band_points", type=int, default=None,
        help=("Number of points calculated on a band segment in the band "
              "structure Gruneisen parameter calculation"))
    parser.add_argument(
        "--bi", "--band-indices", nargs='+', dest="band_indices", default=None,
        help="Band indices where life time is calculated")
    parser.add_argument(
        "--boundary-mfp", "--bmfp", dest="boundary_mfp", type=float,
        default=None,
        help=("Boundary mean free path in micrometre for thermal conductivity "
              "calculation"))
    parser.add_argument(
        "--br", "--bterta", dest="is_bterta", action="store_true",
        default=False,
        help="Calculate thermal conductivity in BTE-RTA")
    if not load_phono3py_yaml:
        parser.add_argument(
            "-c", "--cell", dest="cell_filename", metavar="FILE", default=None,
            help="Read unit cell")
    parser.add_argument(
        "--cf2", "--create-f2", dest="forces_fc2", nargs='+', default=None,
        help="Create FORCES_FC2")
    parser.add_argument(
        "--cf3", "--create-f3", dest="forces_fc3", nargs='+', default=None,
        help="Create FORCES_FC3")
    parser.add_argument(
        "--cfz", "--subtract-forces", dest="forces_fcz", default=None,
        help="Subtract recidual forces from supercell forces")
    parser.add_argument(
        "--cf3-file", "--create-f3-from-file", dest="forces_fc3_file",
        default=None,
        help="Create FORCES_FC3 from file name list")
    parser.add_argument(
        "--cfc", "--compact-fc", dest="is_compact_fc", action="store_true",
        default=False,
        help="Use compact force cosntants")
    parser.add_argument(
        "--cfs", "--create-force-sets", dest="force_sets_mode",
        action="store_true", default=False,
        help="Create phonopy FORCE_SETS from FORCES_FC2")
    parser.add_argument(
        "--cph", "--collective-phonon", dest="solve_collective_phonon",
        action="store_true", default=False,
        help="Solve collective phonons")
    parser.add_argument(
        "--const-ave-pp", dest="const_ave_pp", type=float, default=None,
        help="Set constant averaged ph-ph interaction (Pqj)")
    parser.add_argument(
        "--cutoff-fc3", "--cutoff-fc3-distance", dest="cutoff_fc3_distance",
        type=float, default=None,
        help=("Cutoff distance of third-order force constants. Elements where "
              "any pair of atoms has larger distance than cut-off distance "
              "are set zero."))
    parser.add_argument(
        "--cutoff-freq", "--cutoff-frequency", dest="cutoff_frequency",
        type=float, default=None,
        help="Phonon modes below this frequency are ignored.")
    parser.add_argument(
        "--cutoff-pair", "--cutoff-pair-distance", dest="cutoff_pair_distance",
        type=float, default=None,
        help=("Cutoff distance between pairs of displaced atoms used for "
              "supercell creation with displacements and making third-order "
              "force constants"))
    parser.add_argument(
        "-d", "--disp", dest="is_displacement", action="store_true",
        default=False,
        help="As first stage, get least displacements")
    parser.add_argument(
        "--dim", nargs='+', dest="supercell_dimension", default=None,
        help="Supercell dimension")
    parser.add_argument(
        "--dim-fc2", nargs='+', dest="phonon_supercell_dimension",
        default=None,
        help="Supercell dimension for extra fc2")
    parser.add_argument(
        "--factor", dest="frequency_conversion_factor", type=float,
        default=None,
        help="Frequency unit conversion factor")
    if not load_phono3py_yaml:
        parser.add_argument(
            "--fc2", dest="read_fc2", action="store_true", default=False,
            help="Read second order force constants")
        parser.add_argument(
            "--fc3", dest="read_fc3", action="store_true", default=False,
            help="Read third order force constants")
    parser.add_argument(
        "--fc-calc-opt", "--fc-calculator-options",
        dest="fc_calculator_options", default=None,
        help=("Options for force constants calculator as comma separated "
              "string with the style of key = values"))
    if not fc_symmetry:
        parser.add_argument(
            "--fc-symmetry", "--sym-fc", dest="fc_symmetry",
            action="store_true", default=False,
            help="Symmetrize force constants")
    parser.add_argument(
        "--freq-scale", dest="frequency_scale_factor", type=float,
        default=None,
        help=("Factor multiplied as fc2 * factor^2 and fc3 * factor^2. "
              "Phonon frequency is changed but the contribution from NAC is "
              "not changed."))
    parser.add_argument(
        "--freq-pitch", dest="fpitch", type=float, default=None,
        help="Pitch in frequency for spectrum")
    parser.add_argument(
        "--fs2f2", "--force-sets-to-forces-fc2",
        dest="force_sets_to_forces_fc2_mode", default=False,
        action="store_true", help="Create FORCES_FC2 from FORCE_SETS")
    parser.add_argument(
        "--fst", "--frequency-shift", dest="is_frequency_shift", default=False,
        action="store_true", help="Calculate frequency shifts")
    parser.add_argument(
        "--full-pp", dest="is_full_pp", action="store_true", default=False,
        help=("Calculate full ph-ph interaction for RTA conductivity."
              "This may be activated when full elements of ph-ph interaction "
              "strength are needed, i.e., to calculate average ph-ph "
              "interaction strength."))
    parser.add_argument(
        "--ga", "--grid-addresses", nargs='+', dest="grid_addresses",
        default=None,
        help="Fixed grid addresses where anharmonic properties are calculated")
    parser.add_argument(
        "--gamma-unit-conversion", dest="gamma_unit_conversion", type=float,
        default=None,
        help="Conversion factor for gamma")
    parser.add_argument(
        "--gp", "--grid-points", nargs='+', dest="grid_points", default=None,
        help="Fixed grid points where anharmonic properties are calculated")
    parser.add_argument(
        "--gruneisen", dest="is_gruneisen", action="store_true", default=False,
        help="Calculate phonon Gruneisen parameter")
    parser.add_argument(
        "--gv-delta-q", dest="gv_delta_q", type=float, default=None,
        help="Delta-q distance used for group velocity calculation")
    parser.add_argument(
        "--hdf5-compression", dest="hdf5_compression", default=None,
        help="hdf5 compression filter (default: gzip)")
    if not load_phono3py_yaml:
        parser.add_argument(
            "-i", dest="input_filename", default=None,
            help="Input filename extension")
        parser.add_argument(
            "--io", dest="input_output_filename", default=None,
            help="Input and output filename extension")
    parser.add_argument(
        "--ion-clamped", dest="ion_clamped", action="store_true",
        default=False,
        help=("Atoms are clamped under applied strain in Gruneisen parameter "
              "calculation"))
    parser.add_argument(
        "--ise", dest="is_imag_self_energy", action="store_true",
        default=False,
        help="Calculate imaginary part of self energy")
    parser.add_argument(
        "--isotope", dest="is_isotope", action="store_true", default=False,
        help="Isotope scattering lifetime")
    parser.add_argument(
        "--jdos", dest="is_joint_dos", action="store_true", default=False,
        help="Calculate joint density of states")
    parser.add_argument(
        "--lbte", dest="is_lbte", action="store_true", default=False,
        help="Calculate thermal conductivity LBTE with Chaput's method")
    parser.add_argument(
        "--loglevel", dest="log_level", type=int, default=None,
        help="Log level")
    parser.add_argument(
        "--mass", nargs='+', dest="masses", default=None,
        help="Same as MASS tag")
    parser.add_argument(
        "--md", "--mesh-divisors", nargs='+', dest="mesh_divisors",
        default=None,
        help="Divisors for mesh numbers")
    parser.add_argument(
        "--mesh", nargs='+', dest="mesh_numbers", default=None,
        help="Mesh numbers")
    parser.add_argument(
        "--mv", "--mass-variances", nargs='+', dest="mass_variances",
        default=None,
        help="Mass variance parameters for isotope scattering")
    if not is_nac:
        parser.add_argument(
            "--nac", dest="is_nac", action="store_true", default=False,
            help="Non-analytical term correction")
    parser.add_argument(
        "--nac-method", dest="nac_method", default=None,
        help="Non-analytical term correction method: Wang (default) or Gonze")
    if fc_symmetry:
        parser.add_argument(
            "--no-fc-symmetry", "--no-sym-fc",
            dest="fc_symmetry", action="store_false",
            default=True, help="Do not symmetrize force constants")
    parser.add_argument(
        "--nodiag", dest="is_nodiag", action="store_true", default=False,
        help="Set displacements parallel to axes")
    parser.add_argument(
        "--noks", "--no-kappa-stars", dest="no_kappa_stars",
        action="store_true", default=False,
        help="Deactivate summation of partial kappa at q-stars"),
    parser.add_argument(
        "--nomeshsym", dest="is_nomeshsym", action="store_true", default=False,
        help="No symmetrization of triplets is made.")
    if is_nac:
        parser.add_argument(
            "--nonac", dest="is_nac", action="store_false", default=True,
            help="Non-analytical term correction")
    parser.add_argument(
        "--nosym", dest="is_nosym", action="store_true", default=False,
        help="Symmetry is not imposed.")
    parser.add_argument(
        "--nu", dest="is_N_U", action="store_true", default=False,
        help="Split Gamma into Normal and Umklapp processes")
    parser.add_argument(
        "--num-freq-points", dest="num_frequency_points", type=int,
        default=None,
        help="Number of sampling points for spectrum")
    if not load_phono3py_yaml:
        parser.add_argument(
            "-o", dest="output_filename", default=None,
            help="Output filename extension")
    parser.add_argument(
        "--pa", "--primitive-axis", "--primitive-axes", nargs='+',
        dest="primitive_axes", default=None,
        help="Same as PRIMITIVE_AXES tags")
    parser.add_argument(
        "--pinv-cutoff", dest="pinv_cutoff", type=float, default=None,
        help="Cutoff frequency (THz) for pseudo inversion of collision matrix")
    parser.add_argument(
        "--pinv-solver", dest="pinv_solver", type=int, default=None,
        help="Switch of LBTE pinv solver")
    parser.add_argument(
        "--pm", dest="is_plusminus_displacements", action="store_true",
        default=False,
        help="Set plus minus displacements")
    parser.add_argument(
        "--pp-unit-conversion", dest="pp_unit_conversion", type=float,
        default=None,
        help="Conversion factor for ph-ph interaction")
    parser.add_argument(
        "--qpoints", nargs='+', dest="qpoints", default=None,
        help="Calculate at specified q-points")
    parser.add_argument(
        "--q-direction", nargs='+', dest="nac_q_direction", default=None,
        help="q-vector direction at q->0 for non-analytical term correction")
    parser.add_argument(
        "-q", "--quiet", dest="quiet", action="store_true", default=False,
        help="Print out smallest information")
    parser.add_argument(
        "--read-collision", dest="read_collision", default=None,
        help="Read collision matrix and Gammas from files")
    parser.add_argument(
        "--read-gamma", dest="read_gamma", action="store_true", default=False,
        help="Read Gammas from files")
    parser.add_argument(
        "--read-phonon", dest="read_phonon", action="store_true",
        default=False,
        help="Read phonons from files")
    parser.add_argument(
        "--read-pp", dest="read_pp", action="store_true", default=False,
        help="Read phonon-phonon interaction strength")
    parser.add_argument(
        "--reducible-colmat", dest="is_reducible_collision_matrix",
        action="store_true", default=False,
        help="Solve reducible collision matrix")
    parser.add_argument(
        "--scattering-event-class", dest="scattering_event_class", type=int,
        default=None,
        help=("Scattering event class 1 or 2 to draw imaginary part of self "
              "energy"))
    parser.add_argument(
        "--sigma", nargs='+', dest="sigma", default=None,
        help=("A sigma value or multiple sigma values (separated by space) "
              "for smearing function"))
    parser.add_argument(
        "--sigma-cutoff", dest="sigma_cutoff_width", type=float, default=None,
        help="Cutoff width of smearing function (ratio to sigma value)")
    parser.add_argument(
        "--stp", "--show-num-triplets", dest="show_num_triplets",
        action="store_true", default=False,
        help=("Show reduced number of triplets to be calculated at "
              "specified grid points"))
    if not load_phono3py_yaml:
        parser.add_argument(
            "--sym-fc2", dest="is_symmetrize_fc2", action="store_true",
            default=False,
            help="Symmetrize fc2 by index exchange")
        parser.add_argument(
            "--sym-fc3r", dest="is_symmetrize_fc3_r", action="store_true",
            default=False,
            help="Symmetrize fc3 in real space by index exchange")
        parser.add_argument(
            "--sym-fc3q", dest="is_symmetrize_fc3_q", action="store_true",
            default=False,
            help="Symmetrize fc3 in reciprocal space by index exchange")
    parser.add_argument(
        "--thm", "--tetrahedron-method", dest="is_tetrahedron_method",
        action="store_true", default=False,
        help="Use tetrahedron method.")
    parser.add_argument(
        "--tmax", dest="tmax", default=None,
        help="Maximum calculated temperature")
    parser.add_argument(
        "--tmin", dest="tmin", default=None,
        help="Minimum calculated temperature")
    parser.add_argument(
        "--ts", nargs='+', dest="temperatures", default=None,
        help="Temperatures for damping functions")
    parser.add_argument(
        "--tstep", dest="tstep", default=None,
        help="Calculated temperature step")
    parser.add_argument(
        "--tolerance", dest="symmetry_tolerance", type=float, default=None,
        help="Symmetry tolerance to search")
    parser.add_argument(
        "--uplo", dest="lapack_zheev_uplo", default=None,
        help="Lapack zheev UPLO (default: L)")
    parser.add_argument(
        "-v", "--verbose", dest="verbose", action="store_true", default=False,
        help="Detailed run-time information is displayed")
    parser.add_argument(
        "--wgp", "--write-grid-points", dest="write_grid_points",
        action="store_true", default=False,
        help=("Write grid address of irreducible grid points for specified "
              "mesh numbers to ir_grid_address.yaml"))
    parser.add_argument(
        "--write-collision", dest="write_collision", action="store_true",
        default=False,
        help="Write collision matrix and Gammas to files")
    parser.add_argument(
        "--write-gamma", dest="write_gamma", action="store_true",
        default=False,
        help="Write imag-part of self energy to files")
    parser.add_argument(
        "--write-gamma-detail", "--write_detailed_gamma",
        dest="write_gamma_detail", action="store_true", default=False,
        help="Write out detailed imag-part of self energy")
    parser.add_argument(
        "--write-phonon", dest="write_phonon", action="store_true",
        default=False,
        help="Write all phonons on grid points to files")
    parser.add_argument(
        "--write-pp", dest="write_pp", action="store_true", default=False,
        help="Write phonon-phonon interaction strength")
    parser.add_argument(
        "--write-lbte-solution", dest="write_LBTE_solution",
        action="store_true", default=False,
        help="Write direct solution of LBTE to hdf5 files")
    if load_phono3py_yaml:
        parser.add_argument(
            "filename", nargs='*', help="phono3py.yaml like file")
    else:
        parser.add_argument(
            "filename", nargs='*', help="Phono3py configure file")

    return parser, deprecated



#!/usr/bin/env python

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import numpy as np
from phonopy.file_IO import (
    write_FORCE_SETS, parse_FORCE_SETS, is_file_phonopy_yaml)
from phonopy.units import VaspToTHz, Bohr, Hartree
from phonopy.cui.collect_cell_info import collect_cell_info
from phonopy.cui.create_force_sets import check_number_of_force_files
from phonopy.interface.calculator import (
    get_force_sets, get_default_physical_units, get_interface_mode)
from phonopy.cui.phonopy_argparse import show_deprecated_option_warnings
from phonopy.phonon.band_structure import get_band_qpoints
from phonopy.cui.phonopy_script import (
    store_nac_params, print_end, print_error, print_error_message,
    print_version, file_exists, get_fc_calculator_params)
from phono3py.version import __version__
from phono3py.file_IO import (
    parse_disp_fc2_yaml, parse_disp_fc3_yaml,
    write_disp_fc2_yaml, read_phonon_from_hdf5, write_phonon_to_hdf5,
    parse_FORCES_FC2, write_FORCES_FC2, write_FORCES_FC3,
    get_length_of_first_line, write_fc3_to_hdf5, write_fc2_to_hdf5)
from phono3py.cui.settings import Phono3pyConfParser
from phono3py.cui.load import set_dataset_and_force_constants
from phono3py import Phono3py, Phono3pyJointDos, Phono3pyIsotope
from phono3py.phonon3.gruneisen import run_gruneisen_parameters
from phono3py.phonon3.triplets import get_grid_point_from_address
from phono3py.cui.phono3py_argparse import get_parser
from phono3py.cui.show_log import (
    show_general_settings, show_phono3py_settings, show_phono3py_cells)
from phono3py.cui.triplets_info import write_grid_points, show_num_triplets
from phono3py.cui.create_supercells import create_phono3py_supercells
from phono3py.cui.create_force_constants import create_phono3py_force_constants
from phono3py.interface.phono3py_yaml import Phono3pyYaml


# import logging
# logging.basicConfig()
# logging.getLogger("phono3py.phonon3.fc3").setLevel(level=logging.DEBUG)


# AA is created at http://www.network-science.de/ascii/.
def print_phono3py():
    print("""        _                      _____
  _ __ | |__   ___  _ __   ___|___ / _ __  _   _
 | '_ \| '_ \ / _ \| '_ \ / _ \ |_ \| '_ \| | | |
 | |_) | | | | (_) | | | | (_) |__) | |_) | |_| |
 | .__/|_| |_|\___/|_| |_|\___/____/| .__/ \__, |
 |_|                                |_|    |___/ """)


def finalize_phono3py(phono3py,
                      confs,
                      log_level,
                      displacements_mode=False,
                      filename="phono3py.yaml"):
    """Write phono3py.yaml and then exit.

    Parameters
    ----------
    phono3py : Phono3py
        Phono3py instance.
    confs : dict
        This contains the settings and command options that the user set.
    log_level : int
        Log level. 0 means quiet.
    displacements_mode : Bool
        When True, crystal structure is written in the length unit of
        calculator interface in phono3py_disp.yaml. Otherwise, the
        default unit (angstrom) is used.
    filename : str, optional
        phono3py.yaml is written in this filename.

    """

    if displacements_mode:
        _calculator = phono3py.calculator
    else:
        _calculator = None
    _physical_units = get_default_physical_units(_calculator)

    yaml_settings = {
        'force_sets': False,
        'displacements': displacements_mode
    }

    ph3py_yaml = Phono3pyYaml(configuration=confs,
                              physical_units=_physical_units,
                              settings=yaml_settings)
    ph3py_yaml.set_phonon_info(phono3py)
    ph3py_yaml.calculator = _calculator
    with open(filename, 'w') as w:
        w.write(str(ph3py_yaml))

    if log_level > 0:
        print("")
        print("Summary of calculation was written in \"%s\"." % filename)
        print_end()
    sys.exit(0)


def get_run_mode(settings):
    run_mode = None
    if settings.is_gruneisen:
        run_mode = "gruneisen"
    elif settings.is_joint_dos:
        run_mode = "jdos"
    elif settings.is_isotope and not (settings.is_bterta or settings.is_lbte):
        run_mode = "isotope"
    elif settings.is_imag_self_energy:
        run_mode = "imag_self_energy"
    elif settings.is_frequency_shift:
        run_mode = "frequency_shift"
    elif settings.is_bterta:
        run_mode = "conductivity-RTA"
    elif settings.is_lbte:
        run_mode = "conductivity-LBTE"
    elif settings.create_displacements:
        run_mode = "displacements"
    elif settings.write_phonon:
        run_mode = "phonon"
    return run_mode


def start_phono3py(**argparse_control):
    """Parse arguments and set some basic parameters"""

    parser, deprecated = get_parser(**argparse_control)
    args = parser.parse_args()

    # Log level
    log_level = 1
    if args.verbose:
        log_level = 2
    if args.quiet:
        log_level = 0
    if args.log_level is not None:
        log_level = args.log_level

    # Title
    if log_level:
        print_phono3py()
        print_version(__version__)
        if argparse_control.get('load_phono3py_yaml', False):
            print("Running in phono3py.load mode.")
        print("Python version %d.%d.%d" % sys.version_info[:3])
        import spglib
        print("Spglib version %d.%d.%d" % spglib.get_version())

        if deprecated:
            show_deprecated_option_warnings(deprecated)

    return args, log_level


def read_phono3py_settings(args, argparse_control, log_level):
    """Read phono3py settings"""

    load_phono3py_yaml = argparse_control.get('load_phono3py_yaml', False)

    if len(args.filename) > 0:
        file_exists(args.filename[0], log_level)
        if load_phono3py_yaml:
            phono3py_conf_parser = Phono3pyConfParser(
                args=args, default_settings=argparse_control)
            cell_filename = args.filename[0]
        else:
            if is_file_phonopy_yaml(args.filename[0], keyword='phono3py'):
                phono3py_conf_parser = Phono3pyConfParser(args=args)
                cell_filename = args.filename[0]
            else:
                phono3py_conf_parser = Phono3pyConfParser(
                    filename=args.filename[0], args=args)
                cell_filename = phono3py_conf_parser.settings.cell_filename
    else:
        phono3py_conf_parser = Phono3pyConfParser(args=args)
        cell_filename = phono3py_conf_parser.settings.cell_filename

    confs = phono3py_conf_parser.confs.copy()
    settings = phono3py_conf_parser.settings

    return settings, confs, cell_filename


def create_phono3py_files_then_exit(args,
                                    input_filename,
                                    output_filename,
                                    log_level):
    interface_mode = get_interface_mode(vars(args))

    #####################
    # Create FORCES_FC3 #
    #####################
    if args.forces_fc3 or args.forces_fc3_file:
        if input_filename is None:
            disp_filename = 'disp_fc3.yaml'
        else:
            disp_filename = 'disp_fc3.' + input_filename + '.yaml'
        file_exists(disp_filename, log_level)
        disp_dataset = parse_disp_fc3_yaml(filename=disp_filename)
        if log_level:
            print("Displacement dataset was read from \"%s\"." % disp_filename)
        num_atoms = disp_dataset['natom']
        num_disps = len(disp_dataset['first_atoms'])
        for d1 in disp_dataset['first_atoms']:
            for d2 in d1['second_atoms']:
                if 'included' not in d2 or d2['included']:
                    num_disps += 1
        if args.forces_fc3_file:
            file_exists(args.forces_fc3_file, log_level)
            force_filenames = [x.strip() for x in open(args.forces_fc3_file)]
        else:
            force_filenames = args.forces_fc3

        for filename in force_filenames:
            file_exists(filename, log_level)

        if log_level > 0:
            print("Number of displacements: %d" % num_disps)
            print("Number of supercell files: %d" % len(force_filenames))

        if not check_number_of_force_files(num_disps,
                                           force_filenames,
                                           disp_filename):
            force_sets = []
        else:
            force_sets = get_force_sets(interface_mode,
                                        num_atoms,
                                        num_disps,
                                        force_filenames,
                                        disp_filename=disp_filename,
                                        verbose=(log_level > 0))

        if args.forces_fcz:
            force_filename = args.forces_fcz
            file_exists(force_filename, log_level)
            force_set_zero = get_force_sets(interface_mode,
                                            num_atoms,
                                            1,
                                            [force_filename, ],
                                            verbose=(log_level > 0))[0]
            for fs in force_sets:
                fs -= force_set_zero

            if log_level > 0:
                print("Forces in \'%s\' were subtracted from supercell forces."
                      % force_filename)

        if force_sets:
            write_FORCES_FC3(disp_dataset, force_sets, filename="FORCES_FC3")
            if log_level:
                print("")
                print("%s has been created." % "FORCES_FC3")
                print_end()
            sys.exit(0)
        else:
            if log_level:
                print("")
                print("%s could not be created." % "FORCES_FC3")
                print_error()
            sys.exit(1)

    #####################
    # Create FORCES_FC2 #
    #####################
    if args.forces_fc2:
        if input_filename is None:
            disp_filename = 'disp_fc2.yaml'
        else:
            disp_filename = 'disp_fc2.' + input_filename + '.yaml'
        file_exists(disp_filename, log_level)
        disp_dataset = parse_disp_fc2_yaml(filename=disp_filename)
        if log_level:
            print("Displacement dataset was read from \"%s\"." % disp_filename)
        num_atoms = disp_dataset['natom']
        num_disps = len(disp_dataset['first_atoms'])
        force_filenames = args.forces_fc2
        for filename in force_filenames:
            file_exists(filename, log_level)

        if log_level > 0:
            print("Number of displacements: %d" % num_disps)
            print("Number of supercell files: %d" % len(force_filenames))
        force_sets = get_force_sets(interface_mode,
                                    num_atoms,
                                    num_disps,
                                    force_filenames,
                                    disp_filename,
                                    verbose=(log_level > 0))

        if args.forces_fcz:
            force_filename = args.forces_fcz
            file_exists(force_filename, log_level)
            force_set_zero = get_force_sets(interface_mode,
                                            num_atoms,
                                            1,
                                            [force_filename, ],
                                            verbose=(log_level > 0))[0]
            for fs in force_sets:
                fs -= force_set_zero

            if log_level > 0:
                print("Forces in \'%s\' were subtracted from supercell forces."
                      % force_filename)

        if force_sets:
            write_FORCES_FC2(disp_dataset,
                             forces_fc2=force_sets,
                             filename="FORCES_FC2")
            if log_level:
                print("")
                print("%s has been created." % "FORCES_FC2")
                print_end()
            sys.exit(0)
        else:
            if log_level:
                print("")
                print("%s could not be created." % "FORCES_FC2")
                print_error()
            sys.exit(1)

    if args.force_sets_to_forces_fc2_mode:
        filename = 'FORCE_SETS'
        file_exists(filename, log_level)
        disp_dataset = parse_FORCE_SETS(filename=filename)
        write_FORCES_FC2(disp_dataset)
        write_disp_fc2_yaml(disp_dataset, None)

        if log_level:
            print("FORCES_FC2 and disp_fc2.yaml have been created from "
                  "FORCE_SETS.")
            print_end()
        sys.exit(0)

    #####################################
    # Create FORCE_SETS from FORCES_FC* #
    #####################################
    if args.force_sets_mode:
        if args.phonon_supercell_dimension is not None:
            if input_filename is None:
                disp_filename = 'disp_fc2.yaml'
            else:
                disp_filename = 'disp_fc2.' + input_filename + '.yaml'
            forces_filename = "FORCES_FC2"
        else:
            if input_filename is None:
                disp_filename = 'disp_fc3.yaml'
            else:
                disp_filename = 'disp_fc3.' + input_filename + '.yaml'
            forces_filename = "FORCES_FC3"

        with open(forces_filename, 'r') as f:
            len_first_line = get_length_of_first_line(f)

        if len_first_line == 3:
            file_exists(disp_filename, log_level)
            disp_dataset = parse_disp_fc2_yaml(filename=disp_filename)
            file_exists(forces_filename, log_level)
            parse_FORCES_FC2(disp_dataset, filename=forces_filename)
            if log_level:
                print("Displacement dataset was read from \"%s\"."
                      % disp_filename)
            write_FORCE_SETS(disp_dataset)

            if log_level:
                print("FORCE_SETS has been created.")
                print_end()
        else:
            if log_level:
                print("The file format of %s is already readable by phonopy."
                      % forces_filename)
                print_end()

        sys.exit(0)


def get_input_output_filenames_from_args(args):
    input_filename = args.input_filename
    output_filename = args.output_filename
    if args.input_output_filename is not None:
        input_filename = args.input_output_filename
        output_filename = args.input_output_filename
    return input_filename, output_filename


def get_cell_info(settings, cell_filename, symprec, log_level):
    """calculator interface and crystal structure information"""

    cell_info = collect_cell_info(
        supercell_matrix=settings.supercell_matrix,
        primitive_matrix=settings.primitive_matrix,
        interface_mode=settings.calculator,
        cell_filename=cell_filename,
        chemical_symbols=settings.chemical_symbols,
        phonopy_yaml_cls=Phono3pyYaml,
        symprec=symprec,
        return_dict=True)
    if type(cell_info) is str:
        print_error_message(cell_info)
        if log_level > 0:
            print_error()
        sys.exit(1)

    # cell_info keys
    # ('unitcell', 'supercell_matrix', 'primitive_matrix',
    #  'optional_structure_info', 'interface_mode', 'phonopy_yaml')

    cell_info['phonon_supercell_matrix'] = settings.phonon_supercell_matrix
    ph3py_yaml = cell_info['phonopy_yaml']
    if cell_info['phonon_supercell_matrix'] is None and ph3py_yaml:
        ph_smat = ph3py_yaml.phonon_supercell_matrix
        cell_info['phonon_supercell_matrix'] = ph_smat

    return cell_info


def get_default_values(settings):
    mesh_numbers = settings.mesh_numbers
    grid_points = settings.grid_points
    grid_addresses = settings.grid_addresses
    if grid_addresses is not None:
        grid_points = [get_grid_point_from_address(ga, mesh_numbers)
                       for ga in grid_addresses]

    # Brillouin zone integration: Tetrahedron (default) or smearing method
    sigma = settings.sigma
    if sigma is None:
        sigmas = []
    elif isinstance(sigma, float):
        sigmas = [sigma]
    else:
        sigmas = sigma
    if settings.is_tetrahedron_method:
        sigmas = [None] + sigmas
    if len(sigmas) == 0:
        sigmas = [None]

    if settings.temperatures is None:
        if settings.is_joint_dos:
            temperature_points = None
            temperatures = None
        else:
            t_max = settings.max_temperature
            t_min = settings.min_temperature
            t_step = settings.temperature_step
            temperature_points = [0.0, 300.0]  # For spectra
            temperatures = np.arange(t_min, t_max + float(t_step) / 10, t_step)
    else:
        temperature_points = settings.temperatures  # For spectra
        temperatures = settings.temperatures  # For others

    if settings.frequency_conversion_factor is None:
        frequency_factor_to_THz = VaspToTHz
    else:
        frequency_factor_to_THz = settings.frequency_conversion_factor

    if settings.num_frequency_points is None:
        if settings.frequency_pitch is None:
            num_frequency_points = 201
            frequency_step = None
        else:
            num_frequency_points = None
            frequency_step = settings.frequency_pitch
    else:
        num_frequency_points = settings.num_frequency_points
        frequency_step = None

    if settings.frequency_scale_factor is None:
        frequency_scale_factor = None
    else:
        frequency_scale_factor = settings.frequency_scale_factor

    if settings.cutoff_frequency is None:
        cutoff_frequency = 1e-2
    else:
        cutoff_frequency = settings.cutoff_frequency

    params = {}
    params['grid_points'] = grid_points
    params['sigmas'] = sigmas
    params['temperature_points'] = temperature_points
    params['temperatures'] = temperatures
    params['frequency_factor_to_THz'] = frequency_factor_to_THz
    params['num_frequency_points'] = num_frequency_points
    params['frequency_step'] = frequency_step
    params['frequency_scale_factor'] = frequency_scale_factor
    params['cutoff_frequency'] = cutoff_frequency

    return params


def init_phono3py(settings,
                  cell_info,
                  interface_mode,
                  output_filename,
                  symprec,
                  log_level):
    physical_units = get_default_physical_units(interface_mode)
    distance_to_A = physical_units['distance_to_A']

    # Change unit of lattice parameters to angstrom
    unitcell = cell_info['unitcell'].copy()
    if distance_to_A is not None:
        lattice = unitcell.cell
        lattice *= distance_to_A
        unitcell.cell = lattice

    # updated_settings keys
    # ('grid_points', 'sigmas', 'temperature_points', 'temperatures',
    #  'frequency_factor_to_THz', 'num_frequency_points',
    #  'frequency_step', 'frequency_scale_factor',
    #  'cutoff_frequency')
    updated_settings = get_default_values(settings)

    phono3py = Phono3py(
        unitcell,
        cell_info['supercell_matrix'],
        primitive_matrix=cell_info['primitive_matrix'],
        phonon_supercell_matrix=cell_info['phonon_supercell_matrix'],
        masses=settings.masses,
        mesh=settings.mesh_numbers,
        band_indices=settings.band_indices,
        sigmas=updated_settings['sigmas'],
        sigma_cutoff=settings.sigma_cutoff_width,
        cutoff_frequency=updated_settings['cutoff_frequency'],
        frequency_factor_to_THz=updated_settings['frequency_factor_to_THz'],
        is_symmetry=settings.is_symmetry,
        is_mesh_symmetry=settings.is_mesh_symmetry,
        symmetrize_fc3q=settings.is_symmetrize_fc3_q,
        symprec=symprec,
        calculator=interface_mode,
        log_level=log_level,
        lapack_zheev_uplo=settings.lapack_zheev_uplo)

    return phono3py, updated_settings


def store_force_constants(phono3py,
                          settings,
                          ph3py_yaml,
                          physical_units,
                          input_filename,
                          output_filename,
                          load_phono3py_yaml,
                          log_level):
    if load_phono3py_yaml:
        (fc_calculator,
         fc_calculator_options) = get_fc_calculator_params(settings)

        read_fc = set_dataset_and_force_constants(
            phono3py,
            ph3py_yaml=ph3py_yaml,
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            symmetrize_fc=settings.fc_symmetry,
            is_compact_fc=settings.is_compact_fc,
            log_level=log_level)
        if not read_fc['fc3']:
            write_fc3_to_hdf5(phono3py.fc3,
                              p2s_map=phono3py.primitive.p2s_map,
                              compression=settings.hdf5_compression)
            if log_level:
                print("fc3 was written into \"fc3.hdf5\".")
        if not read_fc['fc2']:
            write_fc2_to_hdf5(phono3py.fc2,
                              p2s_map=phono3py.primitive.p2s_map,
                              physical_unit='eV/angstrom^2',
                              compression=settings.hdf5_compression)
            if log_level:
                print("fc2 was written into \"fc2.hdf5\".")
    else:
        create_phono3py_force_constants(
            phono3py,
            settings,
            ph3py_yaml=ph3py_yaml,
            input_filename=input_filename,
            output_filename=output_filename,
            log_level=log_level)


def run_gruneisen_then_exit(phono3py, settings, output_filename, log_level):
    if (settings.mesh_numbers is None and
        settings.band_paths is None and
        settings.qpoints is None):
        print("An option of --mesh, --band, or --qpoints has to be specified.")
        if log_level:
            print_error()
        sys.exit(1)

    if len(phono3py.fc2) != len(phono3py.fc3):
        print("Supercells used for fc2 and fc3 have to be same.")
        if log_level:
            print_error()
        sys.exit(1)

    if settings.band_paths is not None:
        if settings.band_points is None:
            npoints = 51
        else:
            npoints = settings.band_points
        band_paths = settings.band_paths
        bands = get_band_qpoints(band_paths, npoints=npoints)
    else:
        bands = None

    rotations = phono3py.primitive_symmetry.get_pointgroup_operations()
    run_gruneisen_parameters(
        phono3py.fc2,
        phono3py.fc3,
        phono3py.supercell,
        phono3py.primitive,
        bands,
        phono3py.mesh_numbers,
        rotations,
        settings.qpoints,
        nac_params=phono3py.nac_params,
        nac_q_direction=settings.nac_q_direction,
        ion_clamped=settings.ion_clamped,
        factor=VaspToTHz,
        symprec=phono3py.symmetry.tolerance,
        output_filename=output_filename,
        log_level=log_level)

    if log_level:
        print_end()
    sys.exit(0)


def run_jdos_then_exit(phono3py,
                       settings,
                       updated_settings,
                       output_filename,
                       log_level):
    joint_dos = Phono3pyJointDos(
        phono3py.phonon_supercell,
        phono3py.phonon_primitive,
        phono3py.mesh_numbers,
        phono3py.fc2,
        nac_params=phono3py.nac_params,
        nac_q_direction=settings.nac_q_direction,
        sigmas=updated_settings['sigmas'],
        cutoff_frequency=updated_settings['cutoff_frequency'],
        frequency_step=updated_settings['frequency_step'],
        num_frequency_points=updated_settings['num_frequency_points'],
        temperatures=updated_settings['temperature_points'],
        frequency_factor_to_THz=updated_settings['frequency_factor_to_THz'],
        frequency_scale_factor=updated_settings['frequency_scale_factor'],
        is_mesh_symmetry=settings.is_mesh_symmetry,
        symprec=phono3py.symmetry.tolerance,
        output_filename=output_filename,
        log_level=log_level)

    if log_level > 0:
        dm = joint_dos.dynamical_matrix
        if (dm.is_nac() and dm.nac_method == 'gonze'):
            dm.show_Gonze_nac_message()

    joint_dos.run(updated_settings['grid_points'])

    if log_level:
        print_end()
    sys.exit(0)


def run_isotope_then_exit(phono3py, settings, updated_settings, log_level):
    mass_variances = settings.mass_variances
    if settings.band_indices is not None:
        band_indices = np.hstack(settings.band_indices).astype('intc')
    else:
        band_indices = None
    iso = Phono3pyIsotope(
        phono3py.mesh_numbers,
        phono3py.phonon_primitive,
        mass_variances=mass_variances,
        band_indices=band_indices,
        sigmas=updated_settings['sigmas'],
        frequency_factor_to_THz=updated_settings['frequency_factor_to_THz'],
        symprec=phono3py.symmetry.tolerance,
        cutoff_frequency=settings.cutoff_frequency,
        lapack_zheev_uplo=settings.lapack_zheev_uplo)
    iso.init_dynamical_matrix(
        phono3py.fc2,
        phono3py.phonon_supercell,
        phono3py.phonon_primitive,
        nac_params=phono3py.nac_params,
        frequency_scale_factor=updated_settings['frequency_scale_factor'])
    if log_level > 0:
        dm = iso.dynamical_matrix
        if (dm.is_nac() and dm.nac_method == 'gonze'):
            dm.show_Gonze_nac_message()

    iso.run(updated_settings['grid_points'])
    if log_level:
        print_end()
    sys.exit(0)


def init_phph_interaction(phono3py,
                          settings,
                          updated_settings,
                          input_filename,
                          output_filename,
                          log_level):
    ave_pp = settings.constant_averaged_pp_interaction
    phono3py.init_phph_interaction(
        nac_q_direction=settings.nac_q_direction,
        constant_averaged_interaction=ave_pp,
        frequency_scale_factor=updated_settings['frequency_scale_factor'],
        solve_dynamical_matrices=(not settings.read_phonon))

    if log_level > 0:
        dm = phono3py.dynamical_matrix
        if (dm.is_nac() and dm.nac_method == 'gonze'):
            dm.show_Gonze_nac_message()

    if settings.write_phonon:
        freqs, eigvecs, grid_address = phono3py.get_phonon_data()
        filename = write_phonon_to_hdf5(
            freqs,
            eigvecs,
            grid_address,
            phono3py.mesh_numbers,
            compression=settings.hdf5_compression,
            filename=output_filename)
        if filename:
            if log_level:
                print("Phonons are written into \"%s\"." % filename)
        else:
            print("Writing phonons failed.")
            if log_level:
                print_error()
            sys.exit(1)

    if settings.read_phonon:
        phonons = read_phonon_from_hdf5(phono3py.mesh_numbers,
                                        filename=input_filename,
                                        verbose=(log_level > 0))
        if phonons is None:
            print("Reading phonons failed.")
            if log_level:
                print_error()
            sys.exit(1)

        try:
            phono3py.set_phonon_data(*phonons)
        except RuntimeError:
            if log_level:
                print_error()
            sys.exit(1)


def main(**argparse_control):
    load_phono3py_yaml = argparse_control.get('load_phono3py_yaml', False)

    args, log_level = start_phono3py(**argparse_control)
    physical_units = get_default_physical_units(get_interface_mode(vars(args)))

    if load_phono3py_yaml:
        input_filename = None
        output_filename = None
    else:
        (input_filename,
         output_filename) = get_input_output_filenames_from_args(args)

    create_phono3py_files_then_exit(
        args, input_filename, output_filename, log_level)

    settings, confs, cell_filename = read_phono3py_settings(
        args, argparse_control, log_level)

    # Symmetry tolerance. Distance unit depends on interface.
    if settings.symmetry_tolerance is None:
        symprec = 1e-5
    else:
        symprec = settings.symmetry_tolerance

    cell_info = get_cell_info(settings, cell_filename, symprec, log_level)
    unitcell_filename = cell_info['optional_structure_info'][0]
    interface_mode = cell_info['interface_mode']

    run_mode = get_run_mode(settings)
    if args.write_grid_points:
        run_mode = "write_grid_info"
    elif args.show_num_triplets:
        run_mode = "show_triplets_info"

    # -----------------------------------------------------------------------
    # ----------------- 'args' should not be used below. --------------------
    # -----------------------------------------------------------------------

    ######################################################
    # Create supercells with displacements and then exit #
    ######################################################
    if settings.create_displacements:
        phono3py = create_phono3py_supercells(
            cell_info,
            settings,
            symprec,
            output_filename=output_filename,
            interface_mode=interface_mode,
            log_level=log_level)

        finalize_phono3py(phono3py,
                          confs,
                          log_level,
                          displacements_mode=True,
                          filename="phono3py_disp.yaml")

    #######################
    # Initialize phono3py #
    #######################
    # updated_settings keys
    # ('grid_points', 'sigmas', 'temperature_points', 'temperatures',
    #  'frequency_factor_to_THz', 'num_frequency_points',
    #  'frequency_step', 'frequency_scale_factor',
    #  'cutoff_frequency')
    phono3py, updated_settings = init_phono3py(settings,
                                               cell_info,
                                               interface_mode,
                                               output_filename,
                                               symprec,
                                               log_level)

    #################################################
    # Show phono3py settings and crystal structures #
    #################################################
    if log_level:
        show_general_settings(settings,
                              run_mode,
                              phono3py,
                              unitcell_filename,
                              input_filename,
                              output_filename)

    if log_level > 1:
        show_phono3py_cells(phono3py, settings)
    else:
        print("Spacegroup: %s" %
              phono3py.symmetry.get_international_table())
        print("Use -v option to watch primitive cell, unit cell, "
              "and supercell structures.")

    ##################
    # Check settings #
    ##################
    run_modes_with_mesh = ("conductivity-RTA", "conductivity-LBTE",
                           "imag_self_energy", "jdos", "isotope",
                           "write_grid_info", "show_triplets_info")
    run_modes_with_gp = ("imag_self_energy", "jdos", "isotope")
    if phono3py.mesh_numbers is None and run_mode in run_modes_with_mesh:
        print("")
        print("Mesh numbers have to be specified.")
        print("")
        if log_level:
            print_error()
        sys.exit(1)

    if (run_mode in run_modes_with_gp and
        updated_settings['grid_points'] is None):
        print("")
        print("Grid point(s) has to be specified.")
        print("")
        if log_level:
            print_error()
        sys.exit(1)

    #########################################################
    # Write ir-grid points and grid addresses and then exit #
    #########################################################
    if run_mode == "write_grid_info":
        write_grid_points(phono3py.primitive,
                          phono3py.mesh_numbers,
                          mesh_divs=settings.mesh_divisors,
                          band_indices=settings.band_indices,
                          sigmas=updated_settings['sigmas'],
                          temperatures=updated_settings['temperatures'],
                          coarse_mesh_shifts=settings.coarse_mesh_shifts,
                          is_kappa_star=settings.is_kappa_star,
                          is_lbte=(settings.write_collision or
                                   settings.is_lbte),
                          compression=settings.hdf5_compression,
                          symprec=symprec)

        if log_level:
            print_end()
        sys.exit(0)

    ################################################################
    # Show reduced number of triplets at grid points and then exit #
    ################################################################
    if run_mode == "show_triplets_info":
        show_num_triplets(phono3py.primitive,
                          phono3py.mesh_numbers,
                          mesh_divs=settings.mesh_divisors,
                          band_indices=settings.band_indices,
                          grid_points=updated_settings['grid_points'],
                          coarse_mesh_shifts=settings.coarse_mesh_shifts,
                          is_kappa_star=settings.is_kappa_star,
                          symprec=symprec)

        if log_level:
            print_end()
        sys.exit(0)

    ##################################
    # Non-analytical term correction #
    ##################################
    store_nac_params(phono3py,
                     settings,
                     cell_info['phonopy_yaml'],
                     unitcell_filename,
                     log_level,
                     nac_factor=Hartree * Bohr,
                     load_phonopy_yaml=load_phono3py_yaml)

    ###################
    # Force constants #
    ###################
    store_force_constants(phono3py,
                          settings,
                          cell_info['phonopy_yaml'],
                          physical_units,
                          input_filename,
                          output_filename,
                          load_phono3py_yaml,
                          log_level)

    ############################################
    # Phonon Gruneisen parameter and then exit #
    ############################################
    if settings.is_gruneisen:
        run_gruneisen_then_exit(phono3py, settings, output_filename, log_level)

    #################
    # Show settings #
    #################
    if log_level and run_mode is not None:
        show_phono3py_settings(phono3py, settings, updated_settings, log_level)

    ###########################
    # Joint DOS and then exit #
    ###########################
    if run_mode == "jdos":
        run_jdos_then_exit(phono3py,
                           settings,
                           updated_settings,
                           output_filename,
                           log_level)

    ################################################
    # Mass variances for phonon-isotope scattering #
    ################################################
    if settings.is_isotope and settings.mass_variances is None:
        from phonopy.structure.atoms import isotope_data
        symbols = phono3py.phonon_primitive.symbols
        in_database = True
        for s in set(symbols):
            if s not in isotope_data:
                print("%s is not in the list of isotope databese" % s)
                print("(not implemented).")
                print("Use --mass_variances option.")
                in_database = False
        if not in_database:
            if log_level:
                print_end()
            sys.exit(0)

    #########################################
    # Phonon-isotope lifetime and then exit #
    #########################################
    if run_mode == "isotope":
        run_isotope_then_exit(phono3py, settings, updated_settings, log_level)

    ########################################
    # Initialize phonon-phonon interaction #
    ########################################
    if run_mode is not None:
        init_phph_interaction(phono3py,
                              settings,
                              updated_settings,
                              input_filename,
                              output_filename,
                              log_level)

    #####################################
    # Run imaginary part of self energy #
    #####################################
    if run_mode == "imag_self_energy":
        phono3py.run_imag_self_energy(
            updated_settings['grid_points'],
            frequency_step=updated_settings['frequency_step'],
            num_frequency_points=updated_settings['num_frequency_points'],
            temperatures=updated_settings['temperature_points'],
            scattering_event_class=settings.scattering_event_class,
            write_gamma_detail=settings.write_gamma_detail,
            output_filename=output_filename)
        phono3py.write_imag_self_energy(filename=output_filename)

    ####################################
    # Run lattice thermal conductivity #
    ####################################
    if run_mode == "conductivity-RTA" or run_mode == "conductivity-LBTE":
        phono3py.run_thermal_conductivity(
            is_LBTE=settings.is_lbte,
            temperatures=updated_settings['temperatures'],
            is_isotope=settings.is_isotope,
            mass_variances=settings.mass_variances,
            grid_points=updated_settings['grid_points'],
            boundary_mfp=settings.boundary_mfp,
            solve_collective_phonon=settings.solve_collective_phonon,
            use_ave_pp=settings.use_ave_pp,
            gamma_unit_conversion=settings.gamma_conversion_factor,
            mesh_divisors=settings.mesh_divisors,
            coarse_mesh_shifts=settings.coarse_mesh_shifts,
            is_reducible_collision_matrix=settings.is_reducible_collision_matrix,
            is_kappa_star=settings.is_kappa_star,
            gv_delta_q=settings.group_velocity_delta_q,
            is_full_pp=settings.is_full_pp,
            pinv_cutoff=settings.pinv_cutoff,
            pinv_solver=settings.pinv_solver,
            write_gamma=settings.write_gamma,
            read_gamma=settings.read_gamma,
            write_kappa=True,
            is_N_U=settings.is_N_U,
            write_gamma_detail=settings.write_gamma_detail,
            write_collision=settings.write_collision,
            read_collision=settings.read_collision,
            write_pp=settings.write_pp,
            read_pp=settings.read_pp,
            write_LBTE_solution=settings.write_LBTE_solution,
            compression=settings.hdf5_compression,
            input_filename=input_filename,
            output_filename=output_filename)
    else:
        if log_level:
            print("*" * 15 + " None of ph-ph interaction was calculated. " +
                  "*" * 16)
            print_end()
        sys.exit(0)

    finalize_phono3py(phono3py, confs, log_level)



# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import numpy as np
from phonopy.harmonic.force_constants import (
    show_drift_force_constants,
    symmetrize_force_constants,
    symmetrize_compact_force_constants)
from phonopy.file_IO import get_dataset_type2
from phonopy.cui.phonopy_script import print_error, file_exists
from phonopy.interface.calculator import get_default_physical_units
from phono3py.phonon3.fc3 import show_drift_fc3
from phono3py.file_IO import (
    parse_disp_fc3_yaml, parse_disp_fc2_yaml, parse_FORCES_FC2,
    parse_FORCES_FC3, read_fc3_from_hdf5, read_fc2_from_hdf5,
    write_fc3_to_hdf5, write_fc2_to_hdf5, get_length_of_first_line)
from phono3py.cui.show_log import (
    show_phono3py_force_constants_settings)
from phono3py.phonon3.fc3 import (
    set_permutation_symmetry_fc3, set_translational_invariance_fc3)
from phono3py.interface.phono3py_yaml import Phono3pyYaml


def create_phono3py_force_constants(phono3py,
                                    settings,
                                    ph3py_yaml=None,
                                    input_filename=None,
                                    output_filename=None,
                                    phono3py_yaml_filename=None,
                                    log_level=1):
    if settings.fc_calculator is None:
        symmetrize_fc3r = (settings.is_symmetrize_fc3_r or
                           settings.fc_symmetry)
        symmetrize_fc2 = (settings.is_symmetrize_fc2 or
                          settings.fc_symmetry)
    else:  # Rely on fc calculator the symmetrization of fc.
        symmetrize_fc2 = False
        symmetrize_fc3r = False

    if log_level:
        show_phono3py_force_constants_settings(settings)

    #######
    # fc3 #
    #######
    if (settings.is_joint_dos or
        (settings.is_isotope and
         not (settings.is_bterta or settings.is_lbte)) or
        settings.read_gamma or
        settings.read_pp or
        (not settings.is_bterta and settings.write_phonon) or
        settings.constant_averaged_pp_interaction is not None):
        pass
    else:
        if settings.read_fc3:
            _read_phono3py_fc3(phono3py,
                               symmetrize_fc3r,
                               input_filename,
                               log_level)
        else:  # fc3 from FORCES_FC3 or ph3py_yaml
            _create_phono3py_fc3(phono3py,
                                 ph3py_yaml,
                                 symmetrize_fc3r,
                                 symmetrize_fc2,
                                 input_filename,
                                 output_filename,
                                 settings.is_compact_fc,
                                 settings.cutoff_pair_distance,
                                 settings.fc_calculator,
                                 settings.fc_calculator_options,
                                 log_level)
            if output_filename is None:
                filename = 'fc3.hdf5'
            else:
                filename = 'fc3.' + output_filename + '.hdf5'
            if log_level:
                print("Writing fc3 to \"%s\"." % filename)
            write_fc3_to_hdf5(phono3py.fc3,
                              filename=filename,
                              p2s_map=phono3py.primitive.p2s_map,
                              compression=settings.hdf5_compression)

        cutoff_distance = settings.cutoff_fc3_distance
        if cutoff_distance is not None and cutoff_distance > 0:
            if log_level:
                print("Cutting-off fc3 by zero (cut-off distance: %f)" %
                      cutoff_distance)
            phono3py.cutoff_fc3_by_zero(cutoff_distance)

        if log_level:
            show_drift_fc3(phono3py.fc3, primitive=phono3py.primitive)

    #######
    # fc2 #
    #######
    phonon_primitive = phono3py.phonon_primitive
    p2s_map = phonon_primitive.p2s_map
    if settings.read_fc2:
        _read_phono3py_fc2(phono3py,
                           symmetrize_fc2,
                           input_filename,
                           log_level)
    else:
        if phono3py.phonon_supercell_matrix is None:
            _create_phono3py_fc2(phono3py,
                                 ph3py_yaml,
                                 symmetrize_fc2,
                                 input_filename,
                                 settings.is_compact_fc,
                                 settings.fc_calculator,
                                 settings.fc_calculator_options,
                                 log_level)
        else:
            _create_phono3py_phonon_fc2(phono3py,
                                        ph3py_yaml,
                                        symmetrize_fc2,
                                        input_filename,
                                        settings.is_compact_fc,
                                        settings.fc_calculator,
                                        settings.fc_calculator_options,
                                        log_level)
        if output_filename is None:
            filename = 'fc2.hdf5'
        else:
            filename = 'fc2.' + output_filename + '.hdf5'
        if log_level:
            print("Writing fc2 to \"%s\"." % filename)
        write_fc2_to_hdf5(phono3py.fc2,
                          filename=filename,
                          p2s_map=p2s_map,
                          physical_unit='eV/angstrom^2',
                          compression=settings.hdf5_compression)

    if log_level:
        show_drift_force_constants(phono3py.fc2,
                                   primitive=phonon_primitive,
                                   name='fc2')


def parse_forces(phono3py,
                 ph3py_yaml=None,
                 cutoff_pair_distance=None,
                 force_filename="FORCES_FC3",
                 disp_filename=None,
                 fc_type=None,
                 log_level=0):
    filename_read_from = None

    if fc_type == 'phonon_fc2':
        natom = len(phono3py.phonon_supercell)
    else:
        natom = len(phono3py.supercell)

    # Get dataset from ph3py_yaml. dataset can be None.
    dataset = _extract_datast_from_ph3py_yaml(ph3py_yaml, fc_type)
    if dataset:
        filename_read_from = ph3py_yaml.yaml_filename

    # Try to read FORCES_FC* if type-2 and return dataset.
    # None is returned unless type-2.
    # can emit FileNotFoundError.
    if dataset is not None and not forces_in_dataset(dataset):
        _dataset = _get_type2_dataset(natom,
                                      phono3py.calculator,
                                      filename=force_filename,
                                      log_level=log_level)
        # Do not overwrite dataset when _dataset is None.
        if _dataset:
            filename_read_from = force_filename
            dataset = _dataset

    # Displacement dataset is obtained from disp_filename.
    # can emit FileNotFoundError.
    if dataset is None:
        dataset = _read_disp_fc_yaml(disp_filename, fc_type)
        filename_read_from = disp_filename

    if dataset['natom'] != natom:
        msg = ("Number of atoms in supercell is not consistent with "
               "\"%s\"." % filename_read_from)
        raise RuntimeError(msg)

    if log_level:
        print("Displacement dataset for %s was read from \"%s\"."
              % (fc_type, filename_read_from))

    if cutoff_pair_distance:
        if ('cutoff_distance' not in dataset or
            ('cutoff_distance' in dataset and
             cutoff_pair_distance < dataset['cutoff_distance'])):
            dataset['cutoff_distance'] = cutoff_pair_distance
            if log_level:
                print("Cutoff-pair-distance: %f" % cutoff_pair_distance)

    # Type-1 FORCES_FC*.
    # dataset comes either from disp_fc*.yaml or phono3py*.yaml.
    if not forces_in_dataset(dataset):
        if fc_type == 'phonon_fc2':
            parse_FORCES_FC2(dataset, filename=force_filename)
        else:
            parse_FORCES_FC3(dataset, filename=force_filename)

        if log_level:
            print("Sets of supercell forces were read from \"%s\"."
                  % force_filename)
            sys.stdout.flush()

    _convert_unit_in_dataset(dataset, phono3py.calculator)

    return dataset


def forces_in_dataset(dataset):
    return ('forces' in dataset or
            ('first_atoms' in dataset and
             'forces' in dataset['first_atoms'][0]))


def _read_disp_fc_yaml(disp_filename, fc_type):
    if fc_type == 'phonon_fc2':
        dataset = parse_disp_fc2_yaml(filename=disp_filename)
    else:
        dataset = parse_disp_fc3_yaml(filename=disp_filename)

    return dataset


def _read_phono3py_fc3(phono3py,
                       symmetrize_fc3r,
                       input_filename,
                       log_level):
    if input_filename is None:
        filename = 'fc3.hdf5'
    else:
        filename = 'fc3.' + input_filename + '.hdf5'
    file_exists(filename, log_level)
    if log_level:
        print("Reading fc3 from \"%s\"." % filename)

    p2s_map = phono3py.primitive.p2s_map
    try:
        fc3 = read_fc3_from_hdf5(filename=filename, p2s_map=p2s_map)
    except RuntimeError:
        import traceback
        traceback.print_exc()
        if log_level:
            print_error()
        sys.exit(1)
    num_atom = phono3py.supercell.get_number_of_atoms()
    if fc3.shape[1] != num_atom:
        print("Matrix shape of fc3 doesn't agree with supercell size.")
        if log_level:
            print_error()
        sys.exit(1)

    if symmetrize_fc3r:
        set_translational_invariance_fc3(fc3)
        set_permutation_symmetry_fc3(fc3)

    phono3py.fc3 = fc3


def _read_phono3py_fc2(phono3py,
                       symmetrize_fc2,
                       input_filename,
                       log_level):
    if input_filename is None:
        filename = 'fc2.hdf5'
    else:
        filename = 'fc2.' + input_filename + '.hdf5'
    file_exists(filename, log_level)
    if log_level:
        print("Reading fc2 from \"%s\"." % filename)

    num_atom = phono3py.phonon_supercell.get_number_of_atoms()
    p2s_map = phono3py.phonon_primitive.p2s_map
    try:
        phonon_fc2 = read_fc2_from_hdf5(filename=filename, p2s_map=p2s_map)
    except RuntimeError:
        import traceback
        traceback.print_exc()
        if log_level:
            print_error()
        sys.exit(1)

    if phonon_fc2.shape[1] != num_atom:
        print("Matrix shape of fc2 doesn't agree with supercell size.")
        if log_level:
            print_error()
        sys.exit(1)

    if symmetrize_fc2:
        if phonon_fc2.shape[0] == phonon_fc2.shape[1]:
            symmetrize_force_constants(phonon_fc2)
        else:
            symmetrize_compact_force_constants(phonon_fc2,
                                               phono3py.phonon_primitive)

    phono3py.fc2 = phonon_fc2


def _get_type2_dataset(natom, calculator, filename="FORCES_FC3", log_level=0):
    with open(filename, 'r') as f:
        len_first_line = get_length_of_first_line(f)
        if len_first_line == 6:
            dataset = get_dataset_type2(f, natom)
            if log_level:
                print("%d snapshots were found in %s."
                      % (len(dataset['displacements']), "FORCES_FC3"))
        else:
            dataset = None
    return dataset


def _create_phono3py_fc3(phono3py,
                         ph3py_yaml,
                         symmetrize_fc3r,
                         symmetrize_fc2,
                         input_filename,
                         output_filename,
                         is_compact_fc,
                         cutoff_pair_distance,
                         fc_calculator,
                         fc_calculator_options,
                         log_level):
    """

    Note
    ----
    cutoff_pair_distance is the parameter to determine each displaced
    supercell is included to the computation of fc3. It is assumed that
    cutoff_pair_distance is stored in the step to create sets of
    displacements and the value is stored n the displacement dataset and
    also as the parameter 'included': True or False for each displacement.
    The parameter cutoff_pair_distance here can be used in the step to
    create fc3 by overwriting original cutoff_pair_distance value only
    when the former value is smaller than the later.

    """
    if input_filename is None:
        disp_filename = 'disp_fc3.yaml'
    else:
        disp_filename = 'disp_fc3.' + input_filename + '.yaml'

    _ph3py_yaml = _get_ph3py_yaml(disp_filename, ph3py_yaml)

    try:
        dataset = parse_forces(phono3py,
                               ph3py_yaml=_ph3py_yaml,
                               cutoff_pair_distance=cutoff_pair_distance,
                               force_filename="FORCES_FC3",
                               disp_filename=disp_filename,
                               fc_type='fc3',
                               log_level=log_level)
    except RuntimeError as e:
        # from _parse_forces_type1
        if log_level:
            print(str(e))
            print_error()
        sys.exit(1)
    except FileNotFoundError as e:
        # from _get_type2_dataset
        file_exists(e.filename, log_level)

    phono3py.dataset = dataset
    phono3py.produce_fc3(symmetrize_fc3r=symmetrize_fc3r,
                         is_compact_fc=is_compact_fc,
                         fc_calculator=fc_calculator,
                         fc_calculator_options=fc_calculator_options)


def _create_phono3py_fc2(phono3py,
                         ph3py_yaml,
                         symmetrize_fc2,
                         input_filename,
                         is_compact_fc,
                         fc_calculator,
                         fc_calculator_options,
                         log_level):
    if input_filename is None:
        disp_filename = 'disp_fc3.yaml'
    else:
        disp_filename = 'disp_fc3.' + input_filename + '.yaml'

    _ph3py_yaml = _get_ph3py_yaml(disp_filename, ph3py_yaml)

    try:
        dataset = parse_forces(phono3py,
                               ph3py_yaml=_ph3py_yaml,
                               force_filename="FORCES_FC3",
                               disp_filename=disp_filename,
                               fc_type='fc2',
                               log_level=log_level)
    except RuntimeError as e:
        if log_level:
            print(str(e))
            print_error()
        sys.exit(1)
    except FileNotFoundError as e:
        file_exists(e.filename, log_level)

    phono3py.phonon_dataset = dataset
    phono3py.produce_fc2(
        symmetrize_fc2=symmetrize_fc2,
        is_compact_fc=is_compact_fc,
        fc_calculator=fc_calculator,
        fc_calculator_options=fc_calculator_options)


def _get_ph3py_yaml(disp_filename, ph3py_yaml):
    _ph3py_yaml = ph3py_yaml
    # Try to use phono3py.phonon_dataset when the disp file not found
    if not os.path.isfile(disp_filename):
        disp_filename = None
        if _ph3py_yaml is None and os.path.isfile('phono3py_disp.yaml'):
            _ph3py_yaml = Phono3pyYaml()
            _ph3py_yaml.read('phono3py_disp.yaml')
    return _ph3py_yaml


def _create_phono3py_phonon_fc2(phono3py,
                                ph3py_yaml,
                                symmetrize_fc2,
                                input_filename,
                                is_compact_fc,
                                fc_calculator,
                                fc_calculator_options,
                                log_level):
    if input_filename is None:
        disp_filename = 'disp_fc2.yaml'
    else:
        disp_filename = 'disp_fc2.' + input_filename + '.yaml'

    _ph3py_yaml = _get_ph3py_yaml(disp_filename, ph3py_yaml)

    try:
        dataset = parse_forces(phono3py,
                               ph3py_yaml=_ph3py_yaml,
                               force_filename="FORCES_FC2",
                               disp_filename=disp_filename,
                               fc_type='phonon_fc2',
                               log_level=log_level)
    except RuntimeError as e:
        if log_level:
            print(str(e))
            print_error()
        sys.exit(1)
    except FileNotFoundError as e:
        file_exists(e.filename, log_level)

    phono3py.phonon_dataset = dataset
    phono3py.produce_fc2(
        symmetrize_fc2=symmetrize_fc2,
        is_compact_fc=is_compact_fc,
        fc_calculator=fc_calculator,
        fc_calculator_options=fc_calculator_options)


def _convert_unit_in_dataset(dataset, calculator):
    physical_units = get_default_physical_units(calculator)
    force_to_eVperA = physical_units['force_to_eVperA']
    distance_to_A = physical_units['distance_to_A']

    if 'first_atoms' in dataset:
        for d1 in dataset['first_atoms']:
            if distance_to_A is not None:
                disp = _to_ndarray(d1['displacement'])
                d1['displacement'] = disp * distance_to_A
            if force_to_eVperA is not None and 'forces' in d1:
                forces = _to_ndarray(d1['forces'])
                d1['forces'] = forces * force_to_eVperA
            if 'second_atoms' in d1:
                for d2 in d1['second_atoms']:
                    if distance_to_A is not None:
                        disp = _to_ndarray(d2['displacement'])
                        d2['displacement'] = disp * distance_to_A
                    if force_to_eVperA is not None and 'forces' in d2:
                        forces = _to_ndarray(d2['forces'])
                        d2['forces'] = forces * force_to_eVperA
    else:
        if distance_to_A is not None and 'displacements' in dataset:
            disp = _to_ndarray(dataset['displacements'])
            dataset['displacements'] = disp * distance_to_A
        if force_to_eVperA is not None and 'forces' in dataset:
            forces = _to_ndarray(dataset['forces'])
            dataset['forces'] = forces * force_to_eVperA


def _to_ndarray(array, dtype='double'):
    if type(array) is not np.ndarray:
        return np.array(array, dtype=dtype, order='C')
    else:
        return array


def _extract_datast_from_ph3py_yaml(ph3py_yaml, fc_type):
    dataset = None
    if fc_type == 'phonon_fc2':
        if ph3py_yaml and ph3py_yaml.phonon_dataset is not None:
            # copy dataset
            # otherwise unit conversion can be applied multiple times.
            _ph3py_yaml = Phono3pyYaml()
            _ph3py_yaml.yaml_data = ph3py_yaml.yaml_data
            _ph3py_yaml.parse()
            dataset = _ph3py_yaml.phonon_dataset
    else:
        if ph3py_yaml and ph3py_yaml.dataset is not None:
            # copy dataset
            # otherwise unit conversion can be applied multiple times.
            _ph3py_yaml = Phono3pyYaml()
            _ph3py_yaml.yaml_data = ph3py_yaml.yaml_data
            _ph3py_yaml.parse()
            dataset = _ph3py_yaml.dataset
    return dataset
